// http://intel.github.io/hyperscan/dev-reference/index.html
/*

Unfortunately, we cannot get this test to work to compare performance because:
Hyperscan returns too many matches, for example pattern [A-Z]+ matches the
string abc three times (abc bc c)

I tried to match one pattern at a time using HS_FLAG_SINGLEMATCH.  However,
even with this flag hs_scan() continues to match and does not return until a
number of matches were made.  But to use Hyperscan as a lexer, we need this to
work:

    do
    {
      ctx.id = 0;
      if (hs_scan(database, ctx.subject + ctx.offset, ctx.subject_length - ctx.offset, 0, scratch, eventHandler, &ctx) != HS_SUCCESS)
      {
        exit(1); // error
      }
      // now we have ctx.id that is the pattern ID (pattern index starting at 1)
    }
    while (ctx.id);

The idea here is that hs_scan() invokes eventHandler() that sets ctx.id
and updates ctx.offset for the next match.  We set ctx.id = id to nonzero in
eventHandler(), which is important to determine if a pattern matched and which
pattern matched.

*/

#include <hs/hs.h>
#include <iostream>
#include <fstream>
#include <cassert>
#include "reflex/timer.h"

#define RUNS 10000 // set to 1 to do a quick test

struct context
{
  unsigned int id;
  const char *subject;
  size_t subject_length;
  unsigned long long offset;
  size_t n;
};

static int eventHandler(unsigned int id, unsigned long long from, unsigned long long to, unsigned int flags, void *ctx)
{
  // printf("Match at %llu:%llu\n", from, to);
  ((struct context*)ctx)->offset += to;
  ((struct context*)ctx)->id = id;
  ++((struct context*)ctx)->n;
  return 0;
}

int main()
{
  const char *patterns[56] = {
    "((?:\\s+))",
    "((?:(?:\\Q//\\E)(?:[^\\x0a]|\\\\\\n)*\\n))",
    "((?:(?:\\Q/*\\E)(?:[^\\x2a]|\\*+[^/])*(?:\\Q*/\\E)))",
    "((?:^[\\x09\\x20]*#(?:.|\\\\\\r?\\n)+))",
    "((?:[A-Z_a-z][0-9A-Z_a-z]*))",
    "((?:[\\x2b\\x2d]\\d*(?:\\d|\\.\\d|\\d\\.)\\d*(?:[Ee][\\x2b\\x2d]?\\d+)?[FLfl]?))",
    "((?:[\\x2b\\x2d]?(?:(?:0[0-7]+)|(?:0[Xx][0-9A-Fa-f]+)|(?:0|[1-9]\\d*))(?:[Ll]{0,2}[Uu]?|[Uu][Ll]{0,2})))",
    "((?:L?'(?:\\\\.|[^\\x0a'\\x5c])*'))",
    "((?:L?\"(?:\\\\.|\\\\\\r?\\n|[^\\x0a\"\\x5c])*\"))",
    "((?:\\Q+=\\E))",
    "((?:\\Q++\\E))",
    "((?:\\Q+\\E))",
    "((?:\\Q-=\\E))",
    "((?:\\Q--\\E))",
    "((?:\\Q->\\E))",
    "((?:\\Q-\\E))",
    "((?:\\Q==\\E))",
    "((?:\\Q=\\E))",
    "((?:\\Q<=\\E))",
    "((?:\\Q<<=\\E))",
    "((?:\\Q<<\\E))",
    "((?:\\Q<\\E))",
    "((?:\\Q>=\\E))",
    "((?:\\Q>>=\\E))",
    "((?:\\Q>>\\E))",
    "((?:\\Q>\\E))",
    "((?:\\Q!=\\E))",
    "((?:\\Q!\\E))",
    "((?:\\Q,\\E))",
    "((?:\\Q;\\E))",
    "((?:\\Q...\\E))",
    "((?:\\Q.\\E))",
    "((?:\\Q^=\\E))",
    "((?:\\Q^\\E))",
    "((?:\\Q~\\E))",
    "((?:\\Q*=\\E))",
    "((?:\\Q*\\E))",
    "((?:\\Q/=\\E))",
    "((?:\\Q/\\E))",
    "((?:\\Q%=\\E))",
    "((?:\\Q%\\E))",
    "((?:\\Q&=\\E))",
    "((?:\\Q&&\\E))",
    "((?:\\Q&\\E))",
    "((?:\\Q|=\\E))",
    "((?:\\Q||\\E))",
    "((?:\\Q|\\E))",
    "((?:\\Q(\\E))",
    "((?:\\Q)\\E))",
    "((?:\\Q[\\E))",
    "((?:\\Q]\\E))",
    "((?:\\Q{\\E))",
    "((?:\\Q}\\E))",
    "((?:\\Q::\\E))",
    "((?:\\Q:\\E))",
    "((?:\\Q?\\E))",
  };
  unsigned int flags[56];
  unsigned int ids[56];
  hs_database_t *database;
  hs_compile_error_t *compile_err;

  for (unsigned int i = 0; i < 56; ++i)
  {
    // flags[i] = HS_FLAG_SINGLEMATCH | HS_FLAG_MULTILINE; // see comment
    flags[i] = HS_FLAG_MULTILINE;
    ids[i] = i + 1;
  }

  if (hs_compile_multi(patterns, flags, ids, 56, HS_MODE_BLOCK, NULL, &database, &compile_err) != HS_SUCCESS)
  {
    fprintf(stderr, "ERROR: Unable to compile patterns: %s\n", compile_err->message);
    hs_free_compile_error(compile_err);
    exit(1);
  }

  std::ifstream file("re2test.cpp");
  std::string str((std::istreambuf_iterator<char>(file)), std::istreambuf_iterator<char>());
  file.close();

  struct context ctx;
  ctx.subject = str.c_str();
  ctx.subject_length = str.size();
  ctx.n = 0;

  reflex::timer_type t;
  reflex::timer_start(t);
  for (int run = 0; run < RUNS; ++run)
  {
    ctx.n = 0;
    ctx.offset = 0;
    hs_scratch_t *scratch = NULL;
    if (hs_alloc_scratch(database, &scratch) != HS_SUCCESS)
    {
      fprintf(stderr, "ERROR: Unable to allocate scratch space. Exiting.\n");
      hs_free_database(database);
      exit(1);
    }
    // do // see comment
    {
      // printf("Scanning %llu bytes at offset %llu with Hyperscan\n", ctx.subject_length - ctx.offset, ctx.offset);
      ctx.id = 0;
      if (hs_scan(database, ctx.subject + ctx.offset, ctx.subject_length - ctx.offset, 0, scratch, eventHandler, &ctx) != HS_SUCCESS)
      {
        fprintf(stderr, "ERROR: Unable to scan input buffer. Exiting.\n");
        hs_free_scratch(scratch);
        hs_free_database(database);
        exit(1);
      }
    }
    // while (ctx.id); // see comment
    hs_free_scratch(scratch);
  }
  float dt = reflex::timer_elapsed(t);
  std::cout << ctx.n << " matches, elapsed " << 1000*dt/RUNS << " micro seconds" << std::endl;

  hs_free_database(database);
}
