// http://www.pcre.org/current/doc/html/pcre2demo.html
// this program uses pcre2_dfa_match() instead of pcre2_match() to simulate POSIX matching
// also "optimized" for scanning with options PCRE2_NOTEMPTY and PCRE2_ANCHORED
// c++ -O2 -I/opt/local/include -L/opt/local/lib pcre2posixtest.cpp -lpcre2-8

#define PCRE2_CODE_UNIT_WIDTH 8

#include <pcre2.h>
#include <iostream>
#include <fstream>
#include <cassert>
#include "reflex/timer.h"

#define RUNS 1000

int main()
{
  PCRE2_SPTR pattern = (PCRE2_SPTR)"(?m)(((?:\\s+))|((?:(?:\\Q//\\E)(?:[^\\x0a]|\\\\\\n)*\\n))|((?:(?:\\Q/*\\E)(?:[^\\x2a]|\\*+[^/])*(?:\\Q*/\\E)))|((?:^[\\x09\\x20]*#(?:.|\\\\\\r?\\n)+))|((?:[A-Z_a-z][0-9A-Z_a-z]*))|((?:[\\x2b\\x2d]\\d*(?:\\d|\\.\\d|\\d\\.)\\d*(?:[Ee][\\x2b\\x2d]?\\d+)?[FLfl]?))|((?:[\\x2b\\x2d]?(?:(?:0[0-7]+)|(?:0[Xx][0-9A-Fa-f]+)|(?:0|[1-9]\\d*))(?:[Ll]{0,2}[Uu]?|[Uu][Ll]{0,2})))|((?:L?'(?:\\\\.|[^\\x0a'\\x5c])*'))|((?:L?\"(?:\\\\.|\\\\\\r?\\n|[^\\x0a\"\\x5c])*\"))|((?:\\Q+=\\E))|((?:\\Q++\\E))|((?:\\Q+\\E))|((?:\\Q-=\\E))|((?:\\Q--\\E))|((?:\\Q->\\E))|((?:\\Q-\\E))|((?:\\Q==\\E))|((?:\\Q=\\E))|((?:\\Q<=\\E))|((?:\\Q<<=\\E))|((?:\\Q<<\\E))|((?:\\Q<\\E))|((?:\\Q>=\\E))|((?:\\Q>>=\\E))|((?:\\Q>>\\E))|((?:\\Q>\\E))|((?:\\Q!=\\E))|((?:\\Q!\\E))|((?:\\Q,\\E))|((?:\\Q;\\E))|((?:\\Q...\\E))|((?:\\Q.\\E))|((?:\\Q^=\\E))|((?:\\Q^\\E))|((?:\\Q~\\E))|((?:\\Q*=\\E))|((?:\\Q*\\E))|((?:\\Q/=\\E))|((?:\\Q/\\E))|((?:\\Q%=\\E))|((?:\\Q%\\E))|((?:\\Q&=\\E))|((?:\\Q&&\\E))|((?:\\Q&\\E))|((?:\\Q|=\\E))|((?:\\Q||\\E))|((?:\\Q|\\E))|((?:\\Q(\\E))|((?:\\Q)\\E))|((?:\\Q[\\E))|((?:\\Q]\\E))|((?:\\Q{\\E))|((?:\\Q}\\E))|((?:\\Q::\\E))|((?:\\Q:\\E))|((?:\\Q?\\E)))";

  pcre2_code *re;
  int errornumber;
  PCRE2_SIZE erroroffset;

  re = pcre2_compile(
      pattern,               /* the pattern */
      PCRE2_ZERO_TERMINATED, /* indicates pattern is zero-terminated */
      0,                     /* default options */
      &errornumber,          /* for error number */
      &erroroffset,          /* for error offset */
      NULL);                 /* use default compile context */

  if (re == NULL)
  {
    PCRE2_UCHAR buffer[256];
    pcre2_get_error_message(errornumber, buffer, sizeof(buffer));
    printf("PCRE2 compilation failed at offset %d: %s\n", (int)erroroffset, buffer);
    return 1;
  }

  std::ifstream file("re2test.cpp");
  std::string str((std::istreambuf_iterator<char>(file)), std::istreambuf_iterator<char>());
  file.close();

  PCRE2_SPTR subject = (PCRE2_SPTR)str.c_str();
  size_t subject_length = str.size();
  pcre2_match_data *match_data;
  PCRE2_SIZE wscount = 8192;
  int workspace[wscount];
  PCRE2_SIZE *ovector;
  int rc;

  size_t n = 0;
  reflex::timer_type t;
  reflex::timer_start(t);
  for (int run = 0; run < RUNS; ++run)
  {
    match_data = pcre2_match_data_create_from_pattern(re, NULL);
    ovector = pcre2_get_ovector_pointer(match_data);

    ovector[1] = 0;
    for (n = 0; ; ++n)
    {

      PCRE2_SIZE start_offset = ovector[1];   /* Start at end of previous match */

      /* Run the next matching operation */

      rc = pcre2_dfa_match(
          re,                   /* the compiled pattern */
          subject,              /* the subject string */
          subject_length,       /* the length of the subject */
          start_offset,         /* starting offset in the subject */
          PCRE2_NOTEMPTY | PCRE2_ANCHORED,              /* options */
          match_data,           /* block for storing the result */
          NULL,                 /* use default match context */
          workspace,            /* vector of ints used as working space */
          wscount);             /* number of elements in the vector */

      if (rc == PCRE2_ERROR_NOMATCH)
        break;

      /* Other matching errors are not recoverable. */

      if (rc < 0)
      {
        PCRE2_UCHAR buffer[256];
        pcre2_get_error_message(rc, buffer, sizeof(buffer));
        printf("PCRE2 match failed: %s\n", buffer);
        pcre2_match_data_free(match_data);
        pcre2_code_free(re);
        return 1;
      }

      /* Match succeded */

      // printf("\nMatch succeeded at offset %d\n", (int)ovector[0]);

      /* The match succeeded, but the output vector wasn't big enough. This
         should not happen. */

      if (rc == 0)
        printf("ovector was not big enough for all the captured substrings\n");

    }
  }
  float dt = reflex::timer_elapsed(t);
  std::cout << n << " matches, elapsed " << 1000*dt/RUNS << " micro seconds" << std::endl;

  pcre2_match_data_free(match_data);
  pcre2_code_free(re);
}
