// http://www.pcre.org/current/doc/html/pcre2demo.html
// this program uses pcre2_compile() and pcre2_match() to search input
// c++ -O2 -I/opt/local/include -L/opt/local/lib pcre2test.cpp -lpcre2-8

#define PCRE2_CODE_UNIT_WIDTH 8

#include <pcre2.h>
#include <iostream>
#include <fstream>
#include <cassert>
#include "reflex/timer.h"

#define RUNS 1000

int main()
{
  PCRE2_SPTR pattern = (PCRE2_SPTR)"(?m)(((?:\\s+))|((?:(?:\\Q//\\E)(?:[^\\x0a]|\\\\\\n)*\\n))|((?:(?:\\Q/*\\E)(?:[^\\x2a]|\\*+[^/])*(?:\\Q*/\\E)))|((?:^[\\x09\\x20]*#(?:.|\\\\\\r?\\n)+))|((?:[A-Z_a-z][0-9A-Z_a-z]*))|((?:[\\x2b\\x2d]\\d*(?:\\d|\\.\\d|\\d\\.)\\d*(?:[Ee][\\x2b\\x2d]?\\d+)?[FLfl]?))|((?:[\\x2b\\x2d]?(?:(?:0[0-7]+)|(?:0[Xx][0-9A-Fa-f]+)|(?:0|[1-9]\\d*))(?:[Ll]{0,2}[Uu]?|[Uu][Ll]{0,2})))|((?:L?'(?:\\\\.|[^\\x0a'\\x5c])*'))|((?:L?\"(?:\\\\.|\\\\\\r?\\n|[^\\x0a\"\\x5c])*\"))|((?:\\Q+=\\E))|((?:\\Q++\\E))|((?:\\Q+\\E))|((?:\\Q-=\\E))|((?:\\Q--\\E))|((?:\\Q->\\E))|((?:\\Q-\\E))|((?:\\Q==\\E))|((?:\\Q=\\E))|((?:\\Q<=\\E))|((?:\\Q<<=\\E))|((?:\\Q<<\\E))|((?:\\Q<\\E))|((?:\\Q>=\\E))|((?:\\Q>>=\\E))|((?:\\Q>>\\E))|((?:\\Q>\\E))|((?:\\Q!=\\E))|((?:\\Q!\\E))|((?:\\Q,\\E))|((?:\\Q;\\E))|((?:\\Q...\\E))|((?:\\Q.\\E))|((?:\\Q^=\\E))|((?:\\Q^\\E))|((?:\\Q~\\E))|((?:\\Q*=\\E))|((?:\\Q*\\E))|((?:\\Q/=\\E))|((?:\\Q/\\E))|((?:\\Q%=\\E))|((?:\\Q%\\E))|((?:\\Q&=\\E))|((?:\\Q&&\\E))|((?:\\Q&\\E))|((?:\\Q|=\\E))|((?:\\Q||\\E))|((?:\\Q|\\E))|((?:\\Q(\\E))|((?:\\Q)\\E))|((?:\\Q[\\E))|((?:\\Q]\\E))|((?:\\Q{\\E))|((?:\\Q}\\E))|((?:\\Q::\\E))|((?:\\Q:\\E))|((?:\\Q?\\E)))";

  pcre2_code *re;
  int errornumber;
  PCRE2_SIZE erroroffset;

  re = pcre2_compile(
      pattern,               /* the pattern */
      PCRE2_ZERO_TERMINATED, /* indicates pattern is zero-terminated */
      0,                     /* default options */
      &errornumber,          /* for error number */
      &erroroffset,          /* for error offset */
      NULL);                 /* use default compile context */

  if (re == NULL)
  {
    PCRE2_UCHAR buffer[256];
    pcre2_get_error_message(errornumber, buffer, sizeof(buffer));
    printf("PCRE2 compilation failed at offset %d: %s\n", (int)erroroffset, buffer);
    return 1;
  }

  std::ifstream file("re2test.cpp");
  std::string str((std::istreambuf_iterator<char>(file)), std::istreambuf_iterator<char>());
  file.close();

  PCRE2_SPTR subject = (PCRE2_SPTR)str.c_str();
  size_t subject_length = str.size();
  pcre2_match_data *match_data;
  PCRE2_SIZE *ovector;
  int crlf_is_newline;
  int rc;
  int utf8;
  uint32_t newline;
  uint32_t option_bits;

  size_t n = 0;
  reflex::timer_type t;
  reflex::timer_start(t);
  for (int run = 0; run < RUNS; ++run)
  {
    match_data = pcre2_match_data_create_from_pattern(re, NULL);

    rc = pcre2_match(
        re,                   /* the compiled pattern */
        subject,              /* the subject string */
        subject_length,       /* the length of the subject */
        0,                    /* start at offset 0 in the subject */
        0,                    /* default options */
        match_data,           /* block for storing the result */
        NULL);                /* use default match context */

    /* Matching failed: handle error cases */

    if (rc < 0)
    {
      switch(rc)
      {
        case PCRE2_ERROR_NOMATCH: printf("No match\n"); break;
        default: printf("Matching error %d\n", rc); break;
      }
      pcre2_match_data_free(match_data);   /* Release memory used for the match */
      pcre2_code_free(re);                 /* data and the compiled pattern. */
      return 1;
    }

    ovector = pcre2_get_ovector_pointer(match_data);
    // printf("Match succeeded at offset %d\n", (int)ovector[0]);

    /* The match succeeded, but the output vector wasn't big enough. This
       should not happen. */

    if (rc == 0)
      printf("ovector was not big enough for all the captured substrings\n");

    /* Before running the loop, check for UTF-8 and whether CRLF is a valid newline
       sequence. First, find the options with which the regex was compiled and extract
       the UTF state. */

    (void)pcre2_pattern_info(re, PCRE2_INFO_ALLOPTIONS, &option_bits);
    utf8 = (option_bits & PCRE2_UTF) != 0;

    /* Now find the newline convention and see whether CRLF is a valid newline
       sequence. */

    (void)pcre2_pattern_info(re, PCRE2_INFO_NEWLINE, &newline);
    crlf_is_newline = newline == PCRE2_NEWLINE_ANY ||
      newline == PCRE2_NEWLINE_CRLF ||
      newline == PCRE2_NEWLINE_ANYCRLF;

    /* Loop for second and subsequent matches */

    for (n = 1; ; ++n)
    {

      uint32_t options = 0;                   /* Normally no options */
      PCRE2_SIZE start_offset = ovector[1];   /* Start at end of previous match */

      /* If the previous match was for an empty string, we are finished if we are
         at the end of the subject. Otherwise, arrange to run another match at the
         same point to see if a non-empty match can be found. */

      if (ovector[0] == ovector[1])
      {
        if (ovector[0] == subject_length) break;
        options = PCRE2_NOTEMPTY_ATSTART | PCRE2_ANCHORED;
      }

      /* Run the next matching operation */

      rc = pcre2_match(
          re,                   /* the compiled pattern */
          subject,              /* the subject string */
          subject_length,       /* the length of the subject */
          start_offset,         /* starting offset in the subject */
          options,              /* options */
          match_data,           /* block for storing the result */
          NULL);                /* use default match context */

      /* This time, a result of NOMATCH isn't an error. If the value in "options"
         is zero, it just means we have found all possible matches, so the loop ends.
         Otherwise, it means we have failed to find a non-empty-string match at a
         point where there was a previous empty-string match. In this case, we do what
         Perl does: advance the matching position by one character, and continue. We
         do this by setting the "end of previous match" offset, because that is picked
         up at the top of the loop as the point at which to start again.

         There are two complications: (a) When CRLF is a valid newline sequence, and
         the current position is just before it, advance by an extra byte. (b)
         Otherwise we must ensure that we skip an entire UTF character if we are in
         UTF mode. */

      if (rc == PCRE2_ERROR_NOMATCH)
      {
        if (options == 0) break;                    /* All matches found */
        ovector[1] = start_offset + 1;              /* Advance one code unit */
        if (crlf_is_newline &&                      /* If CRLF is a newline & */
            start_offset < subject_length - 1 &&    /* we are at CRLF, */
            subject[start_offset] == '\r' &&
            subject[start_offset + 1] == '\n')
          ovector[1] += 1;                          /* Advance by one more. */
        else if (utf8)                              /* Otherwise, ensure we */
        {                                         /* advance a whole UTF-8 */
          while (ovector[1] < subject_length)       /* character. */
          {
            if ((subject[ovector[1]] & 0xc0) != 0x80) break;
            ovector[1] += 1;
          }
        }
        continue;    /* Go round the loop again */
      }

      /* Other matching errors are not recoverable. */

      if (rc < 0)
      {
        PCRE2_UCHAR buffer[256];
        pcre2_get_error_message(rc, buffer, sizeof(buffer));
        printf("PCRE2 match failed: %s\n", buffer);
        pcre2_match_data_free(match_data);
        pcre2_code_free(re);
        return 1;
      }

      /* Match succeded */

      // printf("\nMatch succeeded at offset %d\n", (int)ovector[0]);

      /* The match succeeded, but the output vector wasn't big enough. This
         should not happen. */

      if (rc == 0)
        printf("ovector was not big enough for all the captured substrings\n");

    }
    pcre2_match_data_free(match_data);
  }
  float dt = reflex::timer_elapsed(t);
  std::cout << n << " matches, elapsed " << 1000*dt/RUNS << " micro seconds" << std::endl;

  pcre2_code_free(re);
}
